/*
  drv_dsp.c - /dev/dsp driver for fxmp
  Copyright (C) 1999-2001, Bekir Osman Keskin <spectre@portent.net>
  Copyright (C) 2001, Alessandro Gatti <rigel@inet.it>
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <linux/soundcard.h>
#include <sys/ioctl.h>

#include "drv_dsp.h"
#include "fxmp.h"

sshort *soundbuffer;

int dsp_init (fxmp_player *p)
{
  int temp;

  p->driver->dev = open ("/dev/dsp", O_WRONLY, 0);
  if (p->driver->dev == -1)
  {
    printf ("Can't open /dev/dsp\n");
    return 1;
  }

  temp = 1;
  ioctl (p->driver->dev, SNDCTL_DSP_STEREO, &temp);
  temp = 16;
  ioctl (p->driver->dev, SNDCTL_DSP_SAMPLESIZE, &temp);
  temp = FXMP_SAMPLE_RATE;
  ioctl (p->driver->dev, SNDCTL_DSP_SPEED, &temp);
  temp = 16 << 16 | 8;
  ioctl (p->driver->dev, SNDCTL_DSP_SETFRAGMENT, &temp);
  ioctl (p->driver->dev, SNDCTL_DSP_GETBLKSIZE, &temp);
  /*  temp = temp * 8; */

  soundbuffer = (sshort *) malloc (FXMP_BUFFER_SIZE << 2);
  mixer_init ();

  return 0;
}

void dsp_deinit (fxmp_player *p)
{
  close (p->driver->dev);
}

void dsp_soundout (int fd, void *buf, int len)
{
  write (fd, buf, len);
}

void dsp_set_note (fxmp_channel *chan, int note)
{
  chan->period = fxmp_get_period (chan->note, chan->finetune);
  chan->freq = fxmp_get_freq (chan->period);
  chan->sample_inc = (chan->freq << 8) / FXMP_SAMPLE_RATE;
  chan->keyoff = 0;
  chan->pingpong = 0;
  chan->volfade = 32768;
  chan->volfade_dec = chan->insptr->volfade;
  chan->type = chan->smpptr->type & 3;
  chan->sample_pos = 0;
  chan->sample_len = chan->smpptr->length;
  chan->sample_start = 0;

  if (chan->type == FXMP_LOOP_NONE)
    chan->sample_end = chan->sample_len;
  else
    chan->sample_end = chan->smpptr->loopstart + chan->smpptr->looplen;
}


/* dunno if i can find any use of those funcs */
void dsp_set_patch (fxmp_channel *chan, int patch)
{}


void dsp_set_bend (fxmp_channel *chan, int bend)
{}


void dsp_set_volume (fxmp_channel *chan, int volume)
{}


void dsp_set_pan (fxmp_channel *chan, int pan)
{}


void dsp_set_offset (fxmp_channel *chan, int offset)
{}


int dsp_ticker (fxmp_player *p)
{
  audio_buf_info bufinfo;
  slong *buf, nos;
  int st = 0;

  if (p->sleft)
  {
    ioctl (p->driver->dev, SNDCTL_DSP_GETOSPACE, &bufinfo);

    /* nos = bufinfo.bytes >> 2; */
    nos = (bufinfo.fragments * bufinfo.fragsize) >> 2;

    if (nos)
    {
      buf = fxmp_mixer (p, nos);
      fxmp_conv_32to16 (soundbuffer, buf, nos << 2);
      dsp_soundout (p->driver->dev, soundbuffer, nos << 2);

      p->sleft -= nos;
      if (p->sleft <= 0)
      {
        p->sleft += (FXMP_SAMPLE_RATE * 60 / p->module->bpm / 24);
        st = fxmp_ticker (p);
      }
    }
  }
  else
  {
    p->sleft = (FXMP_SAMPLE_RATE * 60 / p->module->bpm / 24);
    st = fxmp_ticker (p);
  }

  return st;
}

fxmp_driver fxmp_drv_dsp =
{
  FXMP_TYPE_SOFTWARE,     /* type */
  0,                      /* dev */
  dsp_init,
  dsp_deinit,
  dsp_set_note,
  dsp_set_patch,
  dsp_set_bend,
  dsp_set_volume,
  dsp_set_pan,
  dsp_set_offset,
  NULL,
  dsp_ticker
};
