/*
  mixer.c - mixer/renderer code
  Copyright (C) 1999-2001, Bekir Osman Keskin <spectre@portent.net>
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "fxmp.h"
#include <stdio.h>
#include <malloc.h>

#ifdef DSP_FILTERS
#include <limits.h>
#include <math.h>

FXMP_FILTER_CALLBACK dsp_output_filter = NULL;
float *source_left;
float *source_right;
float *destination_left;
float *destination_right;
#endif

slong *intbuffer;

#ifdef DSP_FILTERS
void dsp_filter (float *source, float *destination, int length)
{
  /* A simple example, reducing the volume */

  int counter;

  for (counter = 0; counter < length; counter++)
  {
    destination[counter] = source[counter] * 0.1f;
  }
}
#endif

void mixer_init ()
{
  intbuffer = (slong *) malloc (FXMP_BUFFER_SIZE << 2);
  memset (intbuffer, 0, FXMP_BUFFER_SIZE << 2);

#ifdef DSP_FILTERS
  source_left = (float *) malloc ((FXMP_BUFFER_SIZE / 2) * sizeof (float));
  source_right = (float *) malloc ((FXMP_BUFFER_SIZE / 2) * sizeof (float));
  destination_left = (float *) malloc ((FXMP_BUFFER_SIZE / 2) * sizeof (float));
  destination_right = (float *) malloc ((FXMP_BUFFER_SIZE / 2) * sizeof (float));
  dsp_output_filter = (FXMP_FILTER_CALLBACK) &dsp_filter;
#endif
}

void fxmp_conv_32to16 (sshort *out, slong *in, ulong num)
{
  slong smp;
  for (; num--; ++out, ++in)
  {
    smp = *in >> 16;
    smp = smp > 32767 ? 32767 : smp < -32768 ? -32768 : smp;
    *out = (sshort) smp;
  }
}

slong *fxmp_mix_st16norm (slong *out, sshort *samp, ulong count, ulong vr, ulong vl, register slong pos, slong inc)
{
  register sshort smp;

  while (count--)
  {
    smp = samp[pos >> 8];
    *(out++) += smp * vr;
    *(out++) += smp * vl;
    pos += inc;
  }

  return out;
}

slong *fxmp_mixer (fxmp_player *p, int nos)
{
  fxmp_module *xm = p->module;
  ulong volmod, lvolmod, rvolmod;
  fxmp_sample *smpptr;
  fxmp_inst *insptr;
  fxmp_channel *chan;
  uchar volenv = 0;
  sshort pan;
  slong count, tickcount;
  slong *buffer;
  int j;

#ifdef DSP_FILTERS
  int counter;
  int channel_index;
  int intindex;
#endif

  /*  tnos = (FXMP_SAMPLE_RATE * 60 / xm->bpm / 24); */
  memset (intbuffer, 0, nos * 8);

  for (j = 0; j < xm->n_chans; j++)
  {
    chan = &p->channel[j];
    if (!chan->sample_inc) continue;

    pan = (uchar) chan->pan;

    if (chan->inst)
    {
      insptr = &xm->inst[chan->inst - 1];
      smpptr = &insptr->sample[chan->sample_num];

      volmod = chan->volume * p->global_volume;
      volmod = (volmod * chan->volfade) >> 13;

      if (insptr->voltype & 1)
      {
        volenv = insptr->static_volenv[chan->env_volpos];
        volmod = (volmod * volenv) >> 6;
      }

      if (insptr->pantype & 1)
      {
        pan += (insptr->static_panenv[chan->env_panpos] - 32) * (128 - ((pan - 128) & 0x7F)) / 32;
        lvolmod = (volmod * p->pantable[pan].l) >> 5;
        rvolmod = (volmod * p->pantable[pan].r) >> 5;
      }
      else
      {
        lvolmod = rvolmod = volmod;
      }

      buffer = intbuffer;

      for (tickcount = nos; tickcount; )
      {
        count = 1 + ((chan->sample_end << 8) - chan->sample_pos) / chan->sample_inc;
        if (count <= 0) break;
        if (count > tickcount) count = tickcount;

        if (chan->volume > 0)
          buffer = fxmp_mix_st16norm (buffer, smpptr->data, count, lvolmod, rvolmod, chan->sample_pos, chan->sample_inc);

        chan->sample_pos += chan->sample_inc * count;
        tickcount -= count;
        if (tickcount == 0) continue;

        if (chan->type == FXMP_LOOP_NONE)
        {
          tickcount = 0;
          continue;
        }
        else if (chan->type == FXMP_LOOP_FORWARD)
        {
          chan->sample_pos -= (smpptr->looplen) << 8;
        }
        else if (chan->type == FXMP_LOOP_PINGPONG)
        {
          chan->sample_inc = -chan->sample_inc;

          switch (chan->pingpong)
          {
            case 0:
              chan->sample_end = smpptr->loopstart;
              chan->pingpong = 1;
              break;
            case 1:
              chan->sample_end = smpptr->loopstart + smpptr->looplen;
              chan->pingpong = 0;
              break;
          }
        }
      }
    }				/* if (chan->inst) */
  }				/* chanloop */

#ifdef DSP_FILTERS

  if (dsp_output_filter)
  {
    channel_index = 0;
    intindex = 0;

    /* Separate stereo channels */

    for (counter = 0; counter < nos; counter++)
    {
      source_left[channel_index] = (float) ((float) intbuffer[intindex++] / (float) INT_MAX);
      source_right[channel_index] = (float) ((float) intbuffer[intindex++] / (float) INT_MAX);
      channel_index++;
    }

    /* Filter left channel */

    dsp_output_filter (source_left, destination_left, nos);

    /* Filter right channel */

    dsp_output_filter (source_right, destination_right, nos);

    /* Interleave data again */

    channel_index = 0;
    intindex = 0;

    for (counter = 0; counter < nos; counter++)
    {
      intbuffer[intindex++] = (slong) (destination_left[channel_index] * INT_MAX);
      intbuffer[intindex++] = (slong) (destination_right[channel_index] * INT_MAX);
      channel_index++;
    }
  }

#endif /* DSP_FILTERS */

  return intbuffer;
}
